#ifndef __GRAPH_COMMON_H
#define __GRAPH_COMMON_H

// DENNIS added this one
#define OPTIMIZE_TRACE_AFTER_FORCE


#define N_FILE_ENTRIES        251
#define N_FN_ENTRIES         53
#define N_LINE_ENTRIES         37
#define N_INSTR_ENTRIES         793

// DENNIS added
#define NUM_STMT_ENTRIES 1031
#define NUM_TRACE_INSTR_TO_KEEP_ENTRIES 1031
#define NUM_ALT_MAPPING_TO_KEEP_ENTRIES 10007
#define NUM_TRACE_ENTRIES_TABLE_SIZE 10007

#define MAX_x86_INSTR_SIZE    16    // According to ia32 sw dev manual vol 2
#define MIN_LINE_SIZE         16
#define FILE_LEN              256
#define FN_LEN                256

#define INVALID_LINE -1
#define INVALID_UPORT -1
#define VGM_BYTE_INVALID   0xFF
#define VGM_WORD_INVALID   0xFFFFFFFF

#define OPEN_FILE_MAX 10
#define IO_BUF_MAX 64000
#define MAX_USE_PORT 7

#define FLAG_DELIM '%'

// DENNIS added these global variables
extern Bool doOutputTimestamps;
extern Bool doTraceAfterForce;
extern Bool doPdgToTrace;

extern Bool gotToForcedInstance;

#ifdef OPTIMIZE_TRACE_AFTER_FORCE
extern Bool providedAltMappingsToKeep;
#endif

extern Addr VG_(clstk_base);
extern Addr VG_(clstk_end);
extern Addr VG_(client_base);
extern Addr VG_(client_end);
extern Addr VG_(client_mapbase);
extern Addr VG_(client_trampoline_code);
extern Addr VG_(brk_base);	/* start of brk */
extern Addr VG_(brk_limit);	/* current brk */
extern Addr VG_(shadow_base);	/* tool's shadow memory */
extern Addr VG_(shadow_end);
extern Addr VG_(valgrind_base);	/* valgrind's address range */
extern Addr VG_(valgrind_end);

/***************************************/
// DENNIS added these constants to refer to types of locations that are defined or used,
// for outputting properly in the PDG file (each value corresponds to some location)

#define LOCATION_UNDEFINED -500

// for DEFINED values - also uses the register values R_EAX, etc.
#define LOCATION_MEM -110
#define LOCATION_JMP -111

// for USED values - uses the register values R_EAX, etc., to indicate that a used value
// is associated with a REGISTER.  For used values associated with MEMORY, we use the
// *negation* of the register values R_EAX, etc., to indicate the register holds a memory
// *address* (EXCEPTION: since register R_EAX is itself value 0, it has no negation, so
// we need a SPECIAL VALUE to handle R_EAX from memory)
#define LOCATION_MEM_CONST  -120
#define R_EAX_MEM -121

// for DEFINED *and* USED values: flags can be both defined and used
#define LOCATION_FLAG -100

// for FPU_R and FPU_W instructions, mark the defined locations as the FPU stack since the stack
// is affected on these instructions
#define LOCATION_FPU -200


//#define LOCATION_OTHER -110 // for predicate and cd depdendencies that aren't handled right now
/***************************************/

/***************************************/
/* Types of address values */
#define NOT_ADDR    0
#define STACK_ADDR  1
#define HEAP_ADDR   2
#define GLOBAL_ADDR 3
#define OTHER_ADDR  4
/***************************************/

#define INPUT_RELEVANT 0xfffe

//#define CONSIDER_POTENTIAL

//#define INTER_CD

#define TRACK_ARGV

#define TRACK_INPUT

#define VALUE_WANTED

#define BYTE_GRANUITY

//#define CERTIFY

#define TIMESTAMPED

//#define MAPPING

#define MAX_MAPPING_PORT 3
#define MAX_LINK 20 
#define MAX_MAPPING_ENTRY 200

#ifdef MAPPING
typedef struct _val_link val_link;
struct _val_link {
   //Short cnt;
   //Short alloced;
   Int cnt;
   Int uport;
   //Int vals[MAX_MAPPING_PORT];
   Int * expanded[MAX_MAPPING_PORT];
};

typedef struct _val_mapping val_mapping;
struct _val_mapping {
   Int size;
   UInt * indices;
   val_link ** links;
};
#endif

// DENNIS added this struct for list nodes that store integers
typedef struct _intListNode intListNode;
struct _intListNode
{
   Int val;
   intListNode *next;
};

// DENNIS added this struct for list nodes that store Addrs
typedef struct _addrListNode addrListNode;
struct _addrListNode
{
   Addr val;
   addrListNode *next;
};

// DENNIS added this struct for list nodes that store Strings
typedef struct _stringListNode stringListNode;
struct _stringListNode
{
   Char* val;
   stringListNode *next;
};

// Nodes to store heap address info for heap address values
typedef struct _heapAddrNode heapAddrNode;
struct _heapAddrNode
{
   Addr instr; // the program instruction that allocated this heap chunk
   UInt chunkNum; // the chunk number allocated by this instruction (first is #0, next is #1, etc)
   Addr startAddr; // the starting address within the heap for this chunk of space
   UInt size; // size of allocated chunk
   heapAddrNode *next;
};
extern heapAddrNode *heapAddrList; // sorted in increasing order of startAddr

// Nodes to store call stack address info for stack address values
typedef struct _stackNode stackNode;
struct _stackNode
{
   Char* func; // the name of the function associated with this stack node
   UInt instNum; // the instance number of this function invocation (#0 is first invocation, #1 is second invocation, etc)
   Addr baseAddr; // address contained in ESP at function entry
   stackNode *next;
};
extern stackNode *stack; // left end (beginning of list) is top of stack

typedef struct _fnCC fnCC;
struct _fnCC {
   Char*   fn; 
   UInt nextInstanceNumToUse;
   fnCC*   next;
   //lineCC* lines[N_LINE_ENTRIES];
}; 

typedef struct _fileCC fileCC;
struct _fileCC {
   Char*   file;
   fileCC* next;
   fnCC*   fns[N_FN_ENTRIES];
}; 

typedef struct _instr_info instr_info;
   
typedef struct _fun_entry fun_entry;
struct _fun_entry {
   fun_entry * next;
   //Addr orig_addr;
   fnCC * fn; //use fn as the key
   
   fileCC * file;
   //fnCC * fn;
   Addr orig_addr;

   Int orig_sp_offset;
   Int slack;           //the slack between EBP and ESP because push after setting EBP=ESP
   Int alloc_lv;
   //Int used_lv;
   Int cd_available;
   instr_info * first_bb;
};
   

typedef struct _traced_file traced_file;
struct _traced_file {
   Char*   file;
   Char * is_bracket;
   traced_file* next;
}; 
typedef struct _val_stream val_stream;
   

struct _val_stream {
   Int allocated;
   Int used;
   UInt * iis;
   UInt * insts;
   //UInt * visited;   //only used in slicing
};


//storing the potential edges
typedef struct _potential_dep potential_dep;

struct _potential_dep {
   potential_dep * next;
   instr_info * pred_ii;
   Int pred_inst;
};

typedef struct _forward_edge forward_edge;

struct _forward_edge {
   forward_edge * next;
   UInt * ii;
   Int inst;
   UInt value;
};

//This is for relevant slicing. For loading in the potential CD for loads
typedef struct _potential_cd potential_cd;
struct _potential_cd {
     instr_info * load_ii;
     Int pcd_cnt;
     instr_info ** p_pred_bbs;
     instr_info ** p_branches;
};

/*********************************************************************************/
// DENNIS added this struct to store information about mappings to force during execution, for
// when running in FORCE-MAPPING mode
#define MAX_MAPPINGS_TO_FORCE 256
typedef struct _mapping_force mapping_force;
struct _mapping_force
{
   Addr instructionAddr;
   Int instanceNum;
   
   
   
   UInt definedVal;
   Int definedReg;
   Bool hasDefinedFlags;
   UInt definedFlags;
   Bool definedWasForced; // during execution, this is set to True once forcing is done
   UInt definedAddrType; // either not address, stack address, or heap address
   
   // for defined heap values: the instruction goes in heapInstr, the chunk num goes in chunkNum, and
   // the offset goes in definedVal
   Addr heapInstr;
   UInt chunkNum;
   // for defined stack values: the function name goes in funcName, the func instance goes in funcInst,
   // and the offset goes in definedVal
   Char *funcName;
   UInt funcInst;
   
   
   
   
   Int numUses;
   UInt * usedVals;
   Int * usedRegs;
   Bool * usedWasForced; // during execution, this is set to True once forcing is done
   Bool * usedLocationWasReturned; // during instrumentation, is set to True once
                                   // the used location is already instrumented to be forced
                                   
   UInt * usedAddrType;
   
   // for used heap values: the instructions go in heapInstrs, the chunk nums go in chunkNums, and
   // the offsets go in usedVals
   Addr * heapInstrs;
   UInt * chunkNums;
   // for used stack values: the function names go in funcNames, the func instances go in funcInsts,
   // and the offsets go in usedVals
   Char **funcNames;
   UInt * funcInsts;
   
   
   Int currInstance; // this is updated during program execution: it refers to the
                     // current instance number of the x86 instruction corresponding
                     // to instructionAddr that is being or has been exercised.
                     // When currInstance == instanceNum, we need to do this force!
};
/*********************************************************************************/

/*********************************************************************************/
// DENNIS created this structure to hold parameter data that needs to be passed
// to a helper function during program execution that returns a forced value if
// necessary.  The structure was required because there wasn't an instrumentation
// function available that takes as many parameters as I wanted to pass, so I packed
// all parameters into this single struct and passed that as a single parameter

#define MAX_PARAM_STRUCTS_ALLOWED 1024 /* these structs, when allocated, need to
      persist at least until the instrumented code is executed.  So, for now I'll let
      them persist until program termination, and just print an error if we create too
      many of them before the program terminates */
      
typedef struct _force_helper_params force_helper_params;
struct _force_helper_params
{
   Addr x86_addr;
   mapping_force **mappingsToForce;
   Int numMappings;
   Int valueLocation; /* -1 indicates it's the defined value, and any x >= 0 indicates
                         the xth used value from the usedVals array in the mapping_force
                         structure */
                         
   // the following fields are used for doing a force involving a cond/uncond jump instruction pair
   UChar condJumpType;
   UInt condJumpAddr;
   UInt uncondJumpAddr;
};
/*********************************************************************************/


struct _instr_info {
   instr_info * next;
   Addr    orig_addr;    //key
/*
   UChar   instr_size;
   UChar   data_size;
*/

   //Addr     bb_addr;
   //the BBs are created when the control dependence is read in
   instr_info * bb;
   Int bb_id;
   fun_entry * fe;

   //valid for the first instr of a BB
   Int cd_cnt;
   instr_info ** cds; //!=NULL when it is the first instruction in a BB
   instr_info * next_bb;
   instr_info * pred_ii;
   Int pred_id;


   Bool is_predicate;
   Bool true_propagate;
   Bool false_propagate;
   Addr jmp_pc;

   //source code info
   Char *   fn;
   Char *   file;
   Int      line;

   //global id
   Int      id;

   //run time
   Int cur_inst;
   Int lv_idx;

   //dyanmic information
   Int use_port;
   val_stream * uports[MAX_USE_PORT];
   Int usedRegister[MAX_USE_PORT];
   UInt * visited;

   //forward slicing
   forward_edge ** edges;
   UInt * forward_visited;
   Int forward_inst;
#ifdef DEP_DISTANCE
   UInt * back_dist;
   UInt * forward_dist;
#endif

//#ifdef VALUE_WANTED
   Int allocated;
   UInt * values;
   UInt * valuesAddrType;
   
   UInt * instrAllocated; // if this instruction defines a heap/stack address (when stack: it's really ptr to func name!)
   UInt * chunkNum;       // if this instruction defines a heap/stack address (when stack: it's really a func instance num!)
   
   UInt * flags;
   Int definedRegister;
//#endif

   Bool defIsUsedByMainSourceFile;
   UInt nextChunkNumToAllocate; // only if this instruction allocates heap memory

//#ifdef TIMESTAMPED
   //only bb entry has timestamp and control flow info
   Int ts_alloced;
   UInt * tss;
   Addr * cfs;

//#endif


//#ifdef CERTIFIED
   UInt * certified;
//#endif
   Int input_related_alloc;
   UInt * input_related;
#ifdef MAPPING
   val_mapping * mapping;
#endif
   
//for relevant slicing, potential dependences, meaningful for loads
   Int pds_alloced;
   potential_dep ** pds;
   potential_cd * pcd;

};

/* data structure for dynamic switching */

//the set of load instances which potentially depends on the same predicate instance
typedef struct _switch_potential_edge_tail sw_potential_edge_tail;
struct _switch_potential_edge_tail {
     sw_potential_edge_tail * next;
     instr_info * use_ii; 
     Int inst;
     Int ts;
     UInt value;
};

typedef struct _switch_potential_edge_head sw_potential_edge_head;
struct _switch_potential_edge_head {
     sw_potential_edge_head * next;
     sw_potential_edge_head * hash_next;
     instr_info * pred_ii;
     Int inst;
     Int ts;
     sw_potential_edge_tail * tails;
};


typedef struct _io_log io_log;
struct _io_log {
  Int fid;
  Int file_pos;  //what file position the current buf is at
  Int count;
  Int offset;    //the current offset
  Char buf[IO_BUF_MAX];
};

typedef struct _slice_node slice_node;
struct _slice_node {
   slice_node * next;
   instr_info * ii;
   Int inst;
//#ifdef VALUE_WANTED
   UInt value;
   Int step;
//#endif

};

typedef struct _cd_fun_entry cd_fun_entry;
struct _cd_fun_entry {
   Addr entry_addr;
   Char fn[FN_LEN];
   fun_entry * fe;
};

typedef struct _error_node error_node;
struct _error_node {
    Char  file[FILE_LEN];
    Int line;
    Int inst;
};

// DENNIS added these structs
typedef struct _x86InstrStruct x86InstrStruct;
struct _x86InstrStruct // extends hashtable-type struct where the first 2 fields are impt!
{
   x86InstrStruct *next;
   Addr x86Addr; // key
   Char *file;
   Char *function;
   Int line;
   Int currentInstance;
};

typedef struct _SuppressLocation SuppressLocation;
struct _SuppressLocation // extends hashtable-type struct where the first 2 fields are impt!
{
   SuppressLocation *next;
   UInt location; // key
   Bool isSuppressed;
};

typedef struct _suppressPointNode suppressPointNode;
struct _suppressPointNode
{
   Addr instrAddr;
   Int instanceNum;
   suppressPointNode *next;
};
extern suppressPointNode *suppressPointList;

// DENNIS added these globals
extern instr_info *lastInstrExecutedFromMainSourceFile;
extern Int lastInstrInstanceExecutedFromMainSourceFile;
extern Addr smallestInstrAddrExecutedFromMainSourceFile;
extern Addr largestInstrAddrExecutedFromMainSourceFile;
extern Int valgrindMode;

extern void myAssert(Bool condition, char *message);
extern Addr scratchLocation;

#endif
